/*
 *  Copyright (c) 2009 Cyrille Berger <cberger@cberger.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation;
 * either version 2, or (at your option) any later version of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */
#ifndef _TEMPLATE_AST_H_
#define _TEMPLATE_AST_H_

#include <GTLCore/String.h>

namespace GTLCore {
  class Type;
}

namespace OpenCTL {
  class TemplateGenerationContext;
  namespace TemplateAST {
    /**
     * @internal
     * This class represent a node in the AST tree of CTL Templates.
     */
    class Node {
      public:
        virtual ~Node();
        virtual void generate(TemplateGenerationContext* _context) = 0;
      protected:
        /**
         * Utility function that return the string corresponding to the type.
         */
        static GTLCore::String typeToString(const GTLCore::Type* _type);
        static GTLCore::String typeMax(const GTLCore::Type* _type);
        static GTLCore::String typeMin(const GTLCore::Type* _type);
        static GTLCore::String typeUnit(const GTLCore::Type* _type);
    };
    class NodesList : public Node {
      public:
        NodesList( const std::list<Node*>& _nodes);
        virtual ~NodesList();
        virtual void generate(TemplateGenerationContext* _context);
      private:
        std::list<Node* > m_nodes;
    };
    class TextNode : public Node  {
      public:
        TextNode(const GTLCore::String& _text);
        virtual ~TextNode();
        virtual void generate(TemplateGenerationContext* _context);
      private:
        GTLCore::String m_text;
    };
    class OperationNode : public Node {
      public:
        OperationNode( const GTLCore::String& _name, unsigned int inputs, const GTLCore::String& _arg, const GTLCore::String& _optArg);
        virtual ~OperationNode();
        virtual void generate(TemplateGenerationContext* _context);
      private:
        GTLCore::String m_name;
        int m_inputs;
        GTLCore::String m_arg, m_optArg;
    };
    class AllChannelsNode : public Node {
      public:
        enum WhichChannel {
          ColorChannel = 1,
          AlphaChannel = 2,
          AllChannel = ColorChannel | AlphaChannel
        };
      public:
        AllChannelsNode( NodesList* _nodesList, WhichChannel _whichChannel );
        virtual ~AllChannelsNode();
        virtual void generate(TemplateGenerationContext* _context);
      private:
        NodesList* m_nodesList;
        WhichChannel m_whichChannel;
    };
    class AlphaNode : public Node {
      public:
        virtual void generate(TemplateGenerationContext* _context);
    };
    class AlphaMaxNode : public Node {
      public:
        virtual void generate(TemplateGenerationContext* _context);
    };
    class AlphaMinNode : public Node {
      public:
        virtual void generate(TemplateGenerationContext* _context);
    };
    class AlphaUnitNode : public Node {
      public:
        virtual void generate(TemplateGenerationContext* _context);
    };
    class InOutNode : public Node {
      public:
        enum Which {
          In,
          Out
        };
        enum Quantity {
          One,
          All
        };
      public:
        InOutNode(Which _which, Quantity _quantity, unsigned int idx = 0);
      public:
        virtual void generate(TemplateGenerationContext* _context);
      private:
        GTLCore::String m_string;
        GTLCore::String m_idx;
        Quantity m_quantity;
    };
    class VarNode : public Node {
      public:
        VarNode(const GTLCore::String& _name );
      public:
        virtual void generate(TemplateGenerationContext* _context);
      private:
        GTLCore::String m_name;
    };
    class ChannelMaxNode : public Node {
      public:
        virtual void generate(TemplateGenerationContext* _context);
    };
    class ChannelMinNode : public Node {
      public:
        virtual void generate(TemplateGenerationContext* _context);
    };
    class ChannelUnitNode : public Node {
      public:
        virtual void generate(TemplateGenerationContext* _context);
    };
    class ChannelTypeNode : public Node {
      public:
        virtual void generate(TemplateGenerationContext* _context);
    };
    class NamedMaxNode : public Node {
      public:
        NamedMaxNode(const GTLCore::String& _name);
        virtual void generate(TemplateGenerationContext* _context);
      private:
        GTLCore::String m_name;
    };
    class NamedMinNode : public Node {
      public:
        NamedMinNode(const GTLCore::String& _name);
        virtual void generate(TemplateGenerationContext* _context);
      private:
        GTLCore::String m_name;
    };
    class NamedUnitNode : public Node {
      public:
        NamedUnitNode(const GTLCore::String& _name);
        virtual void generate(TemplateGenerationContext* _context);
      private:
        GTLCore::String m_name;
    };
    class NamedTypeNode : public Node {
      public:
        NamedTypeNode(const GTLCore::String& _name);
        virtual void generate(TemplateGenerationContext* _context);
      private:
        GTLCore::String m_name;
    };
  }
}

#endif
